"""
web package for the Ultimate RVC project.

This package contains a module which defines helpful utility functions
for rendering the gradio-based UI of the Ultimate RVC project as well as
submodules which define functions for rendering each tab in that UI.
"""

from ultimate_rvc.core.main import initialize

initialize()
